//$Id:

/*-------------------------------------------------------------------------
 Author: Matthew Galati (magh@lehigh.edu)

 (c) Copyright 2003 Lehigh University. All Rights Reserved.

 This software is licensed under the Common Public License. Please see 
 accompanying file for terms.    
---------------------------------------------------------------------------*/

#ifdef COIN_USE_CLP
#include "OsiClpSolverInterface.hpp"
typedef OsiClpSolverInterface OsiXxxSolverInterface;
#endif

#ifdef COIN_USE_CPX
#include "OsiCpxSolverInterface.hpp"
typedef OsiCpxSolverInterface OsiXxxSolverInterface;
#endif

#ifdef COMPILE_CGL
#include "CglOddHole.hpp"
#include "CglKnapsackCover.hpp"
#include "CglSimpleRounding.hpp"
#include "CglLiftAndProject.hpp"
#include "CglGomory.hpp"
#endif

#include "BCP_lp.hpp"
#include "UFL_lp.hpp"
#include "UFL_cut.hpp"

#include <iostream>
using namespace std;

/*-------------------------------------------------------------------------*/
void UFL_lp::unpack_module_data(BCP_buffer& buf) {
  lp_par.unpack(buf);
  buf.unpack(M).unpack(N).unpack(demand,M);
}

/*-------------------------------------------------------------------------*/
void UFL_lp::pack_cut_algo(const BCP_cut_algo * cut, BCP_buffer & buf){
  const UFL_cut* mcut = dynamic_cast<const UFL_cut*>(cut);
  mcut->pack(buf);
}

/*-------------------------------------------------------------------------*/
BCP_cut_algo * UFL_lp::unpack_cut_algo(BCP_buffer & buf){
  return new UFL_cut(buf);
}

/*-------------------------------------------------------------------------*/
OsiSolverInterface* UFL_lp::initialize_solver_interface(){
  OsiXxxSolverInterface* si = new OsiXxxSolverInterface();
  infinity_ = si->getInfinity();
  return si;
}

/*-------------------------------------------------------------------------*/
void UFL_lp::modify_lp_parameters(OsiSolverInterface* lp, 
				  bool in_strong_branching){
}

/*-------------------------------------------------------------------------*/
void UFL_lp::display_lp_solution(const BCP_lp_result& lpres,
				 const BCP_vec<BCP_var*>& vars,
				 const BCP_vec<BCP_cut*>& cuts,
				 const bool final_lp_solution){

  //if(!final_lp_solution) return;

  const double * solution = lpres.x();
  const double tolerance = lp_par.entry(UFL_lp_par::UFL_tolerance);

  cout.setf(ios::fixed);
  cout.precision(5);
  int c = 0, i, j;
  for(i = 0; i < M; i++){
    for(j = 0; j < N; j++){
      if(solution[c] > tolerance){
	cout << "c : " << c << " x[" << i << "," << j << "] = " 
	     << solution[c] << endl;
      }
      c++;
    }
  }
  for(j = 0; j < N; j++){
    if(solution[c] > tolerance){
      cout << "c : " << c << " y[" << j << "] = " 
	   << solution[c] << endl;
    }
    c++;
  }

}

/*-------------------------------------------------------------------------*/
void UFL_lp::generate_cuts_in_lp(const BCP_lp_result& lpres,
				 const BCP_vec<BCP_var*>& vars,
				 const BCP_vec<BCP_cut*>& cuts,
				 BCP_vec<BCP_cut*>& new_cuts,
				 BCP_vec<BCP_row*>& new_rows){

  /* 
     Look for disaggregated logical constraints:
     x_ij - (d_j * y_j) <= 0, for i in M (client), j in N (depot)
  */

  const double tolerance = lp_par.entry(UFL_lp_par::UFL_tolerance);
  const double* sol      = lpres.x();
  
  int i, j, xind, yind;
  for(i = 0; i < M; i++){
    for(j = 0; j < N; j++){
      xind = xindex(i,j);
      yind = yindex(j);
      if(sol[xind] - (demand[i] * sol[yind]) > tolerance)
	new_cuts.push_back(new UFL_cut(i,j));				
    }
  }

#if 0
  //MVG: not done yet
 
  std::vector<CglCutGenerator*> cgs;
  if (lp_par.entry(UFL_lp_par::UFL_KnapsackCover))
    cgs.push_back(new CglKnapsackCover);
  if (lp_par.entry(UFL_lp_par::UFL_LiftAndProject))
    cgs.push_back(new CglLiftAndProject);
  if (lp_par.entry(UFL_lp_par::UFL_SimpleRounding))
    cgs.push_back(new CglSimpleRounding);
  if (lp_par.entry(UFL_lp_par::UFL_OddHole))
    cgs.push_back(new CglOddHole);
  if (lp_par.entry(UFL_lp_par::UFL_Gomory)){
    CglGomory *gomory = new CglGomory;
    gomory->setLimit(100);
    cgs.push_back(gomory);
  }

  OsiSolverInterface* si = getLpProblemPointer()->lp_solver;
  for (i = vars.size() - 1; i >= 0; --i) {
    if (vars[i]->var_type() != BCP_ContinuousVar) { // integer or binary
      si->setInteger(i);
    }
  }
  OsiCuts cutlist;
  for (i = cgs.size() - 1; i >= 0; --i) {
    cgs[i]->generateCuts(*si, cutlist);
    cout << "UFL-CGL cutlist.sizeRowCuts() " << cutlist.sizeRowCuts() << endl;
  }
  printf("UFL: found %i algo cuts.\n", cutlist.sizeRowCuts());
#endif

}

/*-------------------------------------------------------------------------*/
void UFL_lp::cuts_to_rows(const BCP_vec<BCP_var*>& vars,
			  BCP_vec<BCP_cut*>& cuts,
			  BCP_vec<BCP_row*>& rows,
			  const BCP_lp_result& lpres,
			  BCP_object_origin origin, bool allow_multiple){
  
  const int cutnum = cuts.size();
  rows.reserve(cutnum);

  for (int c = 0; c < cutnum; ++c) {
    const UFL_cut* mcut = dynamic_cast<const UFL_cut*>(cuts[c]);
    if(mcut != 0){
      CoinPackedVector cut;
      cut.insert(xindex(mcut->i,mcut->j), 1.0);
      cut.insert(yindex(mcut->j), -1.0 * demand[mcut->i]);
      rows.push_back(new BCP_row(cut, -1.0 * infinity_, 0.0));
    }
  }
}
