// $Id: UFL_tm.cpp,v 1.2 2003/12/03 01:46:35 magh Exp $

/* ------------------------------------------------------------------------
 Author: Matthew Galati (magh@lehigh.edu)

 (c) Copyright 2003 Lehigh University. All Rights Reserved.

 This software is licensed under the Common Public License. Please see 
 accompanying file for terms.    
---------------------------------------------------------------------------*/

#include "UFL_tm.hpp"
#include "UFL_cut.hpp"
#include "CoinHelperFunctions.hpp"

#include <fstream>
#include <iostream>
using namespace std;

/*-------------------------------------------------------------------------*/
UFL_tm::~UFL_tm(){
  if(trans_cost != 0) { delete [] trans_cost; trans_cost = 0; }
  if(fixed_cost != 0) { delete [] fixed_cost; fixed_cost = 0; }
  if(demand != 0)     { delete [] demand; demand = 0; }
}

/*-------------------------------------------------------------------------*/
void UFL_tm::pack_module_data(BCP_buffer& buf, BCP_process_t ptype){

  switch(ptype){
  case BCP_ProcessType_Any:
  case BCP_ProcessType_TM:
    abort();
  case BCP_ProcessType_CG:
  case BCP_ProcessType_VG:
  case BCP_ProcessType_CP:
  case BCP_ProcessType_VP:  
    break;
  case BCP_ProcessType_LP:
    lp_par.pack(buf);
    break;
  }
  buf.pack(M).pack(N).pack(demand,M);
}

/*-------------------------------------------------------------------------*/
void UFL_tm::pack_cut_algo(const BCP_cut_algo* cut, BCP_buffer& buf){
  const UFL_cut* mcut = dynamic_cast<const UFL_cut*>(cut);
  mcut->pack(buf);
}

/*-------------------------------------------------------------------------*/
BCP_cut_algo* UFL_tm::unpack_cut_algo(BCP_buffer& buf){
  return new UFL_cut(buf);	
}

/*-------------------------------------------------------------------------*/
void UFL_tm::initialize_core(BCP_vec<BCP_var_core*>& vars,
			     BCP_vec<BCP_cut_core*>& cuts,
			     BCP_lp_relax*& lp_relax) {

  /* Read the problem instance. */

  BCP_string dfile = tm_par.entry(UFL_tm_par::UFL_datafile);  
  read_data(dfile.c_str());

  /*
    Next, we must define the columns (variables) in the problem. The 
    functions xindex(i,j) and yindex(j) define the order of the 
    columns. First we list the x indices (i = 0, ..., M-1, j = 0, 
    ..., N-1), then the y indices (j = 0, ..., N-1). We will also store 
    the indices of those columns which are integral (the y variables).
  */
  int n_cols = (M * N) + N;
  double * objective    = new double[n_cols];//the objective coefficients
  double * col_lb       = new double[n_cols];//the column lower bounds
  double * col_ub       = new double[n_cols];//the column upper bounds
  vars.reserve(n_cols);

  /*
    Here we use some COIN helper functions which are defined in:
    Coin/include/CoinHelperFunctions.hpp
  */
  CoinFillN(col_lb, n_cols, 0.0);

  int i, j, index;
  for(i = 0; i < M; i++){
    for(j = 0; j < N; j++){
      index            = xindex(i,j);
      objective[index] = trans_cost[index] / demand[i];
      col_ub[index]    = demand[i];
      vars.push_back(new BCP_var_core(BCP_ContinuousVar, objective[index],
				      0.0, col_ub[index]));
    }
  }

  CoinFillN(col_ub + (M*N), N, 1.0);
  CoinDisjointCopyN(fixed_cost, N, objective + (M * N));
  for(j = 0; j < N; j++){
    vars.push_back(new BCP_var_core(BCP_BinaryVar,fixed_cost[j], 
				    0.0, 1.0));
  }

  /*
    Next we must define the set of rows (constraints). We will store
    the information in a BCP_lp_relax which is derived from 
    CoinPackedMatrix.
  */

  int n_rows = M + N;
  double * row_lb = new double[n_rows]; //the row lower bounds
  double * row_ub = new double[n_rows]; //the row upper bounds
  cuts.reserve(n_rows);

  /*
    There are M demand constraints  (= d[i]), LB[i] = UB[i] = d[i]
    There are N linking constraints (<= 0.0), LB[i] = -infinity, UB = 0.0
    
    Since different solvers define infinity differently, we use the 
    OSI access function getInfinity() to fetch the proper value. ??
  */

  CoinDisjointCopyN(demand, M, row_lb);
  CoinDisjointCopyN(demand, M, row_ub);
  CoinFillN(row_lb + M, N, -1 * infinity_);
  CoinFillN(row_ub + M, N, 0.0);
  
  CoinPackedMatrix * matrix =  new CoinPackedMatrix(false,0,0);
  matrix->setDimensions(0, n_cols);
  
  //demand constraints: 
  //sum{j in N} x[i][j] = d[i], for i in M
  for(i = 0; i < M; i++){
    CoinPackedVector row;
    for(j = 0; j < N; j++)
      row.insert(xindex(i,j), 1.0);
    matrix->appendRow(row);
    cuts.push_back(new BCP_cut_core(demand[i],demand[i]));
  }

  //linking constraints: 
  //sum{i in M} x[i][j] - total_demand * y[j] <= 0, for j in N
  for(j = 0; j < N; j++){
    CoinPackedVector row;
    row.insert(yindex(j), -1.0 * total_demand);
    for(i = 0; i < M; i++)
      row.insert(xindex(i,j), 1.0);
    matrix->appendRow(row);
    cuts.push_back(new BCP_cut_core(-1 * infinity_, 0.0));
  }  

  lp_relax = new BCP_lp_relax();
  lp_relax->assign(*matrix, objective, col_lb, col_ub, row_lb, row_ub);

  /*
    Now free up the memory. Notice, the difference between using OSI's
    loadProblem vs assignProblem. With assignProblem, OSI assumes 
    ownership of the arguments and deletes the memory for the user.
  */

#if 0
  //if(matrix != 0)      { delete matrix; matrix = 0; }
  if(objective != 0)   { delete [] objective; objective = 0; }
  if(col_lb != 0)      { delete [] col_lb; col_lb = 0; }
  if(col_ub != 0)      { delete [] col_ub; col_ub = 0; }
  if(row_lb != 0)      { delete [] row_lb; row_lb = 0; }
  if(row_ub != 0)      { delete [] row_ub; row_ub = 0; }
  //
#endif
}

/*-------------------------------------------------------------------------*/
void UFL_tm::read_data(const char* datafile) {

  /*
    http://mscmga.ms.ic.ac.uk/jeb/orlib/uncapinfo.html

    The format of these data files is:
    number of potential warehouse locations (n), number of customers (m)
    for each potential warehouse location j (j = 1,...,n): 
    capacity, fixed cost
    for each customer i (i = 1,...,m): 
    demand, 
    for each warehouse j (j = 1,...,n)
    cost of allocating all of the demand of i to warehouse j
  */

  ifstream is(datafile);
  if (!is) {
    cout << "ERROR I/O: Failure to open: " << datafile << endl;
    abort();
  }

  //number of depots,clients 
  is >> N >> M;
  
  trans_cost = new double[M * N];
  fixed_cost = new double[N];
  demand     = new double[M];

  string dummy; //we are not using the capacity information
  int i, j;

  for(j = 0; j < N; j++)
    is >> dummy >> fixed_cost[j];

  for(i = 0; i < M; i++){
    is >> demand[i];
    total_demand += demand[i];
    for(j = 0; j < N; j++){
      int index = xindex(i,j);
      is >> trans_cost[index];
    }
  }

  is.close();
}
